"""Tests for src/websnap/logic.py"""

import pytest

from tests.helpers import get_section_config, write_json_config
from websnap.logger import get_custom_logger
from websnap.logic import terminate_program, get_url_content, is_min_size_kb


@pytest.fixture
def log_logic(log_config_model):
    return get_custom_logger(name="websnap_logic", config=log_config_model)


@pytest.fixture
def url_content_pypi_websnap(log_logic):
    return get_url_content(
        "https://pypi.org/pypi/websnap/json", "pypi-websnap", log_logic
    )


@pytest.mark.parametrize(
    "url, section, expected",
    [
        ("https://httpbin.org/status/400", "error-response", None),
        ("https://pypi.org/pypi/websnap/json", "pypi-websnap", bytes),
    ],
)
def test_get_url_content(url, section, expected, log_logic):

    result = get_url_content(url, section, log_logic)

    if expected is None:
        assert result is None
    else:
        assert isinstance(result, expected)


@pytest.mark.parametrize(
    "min_size_kb, section, expected",
    [
        (1, "pypi-websnap", True),
        (50000, "pypi-websnap", False),
    ],
)
def test_is_min_size_kb(
    min_size_kb, section, expected, log_logic, url_content_pypi_websnap
):

    result = is_min_size_kb(
        url_content=url_content_pypi_websnap,
        min_size_kb=min_size_kb,
        section=section,
        log=log_logic,
    )
    assert result == expected


def test_terminate_program():
    assert terminate_program(False) is None


def test_terminate_program_exit():
    with pytest.raises(SystemExit):
        terminate_program(True)



"""Tests configuration with custom argument and shared websnap config fixtures."""

import json
import os

import pytest

from tests.helpers import (
    get_section_config,
    write_json_config,
    get_s3_config,
    get_s3_config_invalid,
)
from websnap.validators import get_config_parser, validate_log_config


def pytest_addoption(parser):
    parser.addoption(
        "--s3-config",
        action="store",
        default=None,
        help="Optional configuration file used to test websnap S3 functionality",
    )


def pytest_configure(config):

    s3_config = config.getoption("--s3-config")

    if s3_config is None:
        return

    if not os.path.exists(s3_config):
        raise pytest.UsageError(f"S3 config '{s3_config}' does not exist")

    with open(s3_config) as config_file:
        s3_config_dict = json.load(config_file)

    for section in s3_config_dict:
        if section == "DEFAULT":
            continue
        for setting in ["url", "bucket", "key"]:
            if setting not in s3_config_dict[section]:
                raise pytest.UsageError(
                    f"S3 config '{s3_config}' section '{section}' is missing required "
                    f"setting '{setting}'"
                )


@pytest.fixture
def s3_config(request):

    s3_config_path = request.config.getoption("--s3-config")

    if not s3_config_path:
        return None

    return s3_config_path


@pytest.fixture
def config_basic(tmp_path):
    file_name = "output_basic.json"
    section_config = get_section_config(tmp_path, file_name)
    config_path = f"{str(tmp_path)}/config_basic.json"

    write_json_config(config_path, section_config)

    return config_path, tmp_path, file_name


@pytest.fixture
def config_parser_basic(config_basic):
    return get_config_parser(config_basic[0])


@pytest.fixture
def config_min_size_kb(tmp_path):
    default_config = {"DEFAULT": {"min_size_kb": 1}}
    file_name = "output_min_size_kb.json"
    section_config = get_section_config(tmp_path, file_name)
    conf_dict = default_config | section_config

    config_path = f"{str(tmp_path)}/config_min_size_kb.json"
    write_json_config(config_path, conf_dict)

    return config_path, tmp_path, file_name


@pytest.fixture
def config_log(tmp_path):
    log_config = {
        "DEFAULT": {
            "log_when": "midnight",
            "log_interval": 2,
            "log_backup_count": 2,
        }
    }
    file_name = "output_log.json"
    section_config = get_section_config(tmp_path, file_name)
    conf_dict = log_config | section_config

    config_path = f"{str(tmp_path)}/config_log.json"
    write_json_config(config_path, conf_dict)

    return config_path, tmp_path, file_name


@pytest.fixture
def config_parser_log(config_log):
    return get_config_parser(config_log[0])


@pytest.fixture
def log_config_model(config_parser_log):
    return validate_log_config(config_parser_log)


@pytest.fixture
def config_s3(tmp_path):
    s3_config = get_s3_config()
    config_path = f"{str(tmp_path)}/config_s3.json"
    write_json_config(config_path, s3_config)
    return config_path, tmp_path


@pytest.fixture
def config_parser_s3(config_s3):
    return get_config_parser(config_s3[0])


@pytest.fixture
def config_s3_invalid(tmp_path):
    s3_conf = get_s3_config_invalid()
    config_path = f"{str(tmp_path)}/config_s3_invalid.json"
    write_json_config(config_path, s3_conf)
    return config_path, tmp_path


@pytest.fixture
def config_parser_s3_invalid(config_s3_invalid):
    return get_config_parser(config_s3_invalid[0])



"""
Function websnap() downloads files from URLs and uploads them to S3 bucket.
Also supports writing downloaded files to local machine.
"""

import logging
import time

from websnap.constants import TIMEOUT
from websnap.validators import (
    get_config_parser,
    validate_log_config,
    validate_s3_config,
    validate_min_size_kb,
    validate_positive_int_args,
)
from websnap.logger import get_custom_logger
from websnap.logic import (
    write_urls_locally,
    write_urls_to_s3,
    sleep_until_next_iteration,
)

__all__ = ["websnap"]

LOGGER_NAME = "websnap"


def websnap(
    config: str = "config.ini",
    log_level: str = "INFO",
    file_logs: bool = False,
    s3_uploader: bool = False,
    backup_s3_count: int | None = None,
    timeout: int = TIMEOUT,
    early_exit: bool = False,
    repeat_minutes: int | None = None,
    section_config: str | None = None,
) -> None:
    """
    Copies files hosted at URLs in config and then uploads them
    to S3 bucket or local machine.

    Args:
        config: Path to .ini or .json configuration file.
        log_level: Level to use for logging.
        file_logs: If True then implements rotating file logs.
        s3_uploader: If True then uploads files to S3 bucket.
        backup_s3_count: Copy and backup S3 objects in each config section
            <backup_s3_count> times,
            remove object with the oldest last modified timestamp.
            If integer passed then it must be a positive integer.
            If omitted then default value is None and objects are not copied.
        timeout: Number of seconds to wait for response for each HTTP request.
            If integer passed then it must be a positive integer.
        early_exit: If True then terminates program immediately after error occurs.
            Default value is False.
            If False then only logs error and continues execution.
        repeat_minutes: Run websnap continuously every <repeat> minutes
               If integer passed then it must be a positive integer.
               If omitted then default value is None and websnap will not repeat.
        section_config: File or URL to obtain additional configuration sections.
                If omitted then default value is None and only config specified in
                'config' argument is used.
                Cannot be used to assign DEFAULT values in the config (DEFAULT
                values must be assigned in config specified by 'config' argument).
                Currently only supports JSON config and can only be used if 'config'
                argument is also a JSON file.
                Duplicate sections will overwrite values with the same section
                passed in the `config` argument.
    """
    # Validate integer arguments
    try:
        validate_positive_int_args(timeout, backup_s3_count, repeat_minutes)
    except Exception as e:
        raise e

    # Validate configuration
    try:
        conf_parser = get_config_parser(config, section_config, timeout)
        conf_log = validate_log_config(conf_parser)
        min_size_kb = validate_min_size_kb(conf_parser)
    except Exception as e:
        raise e

    # Validate custom log
    log = get_custom_logger(
        name=LOGGER_NAME,
        level=log_level,
        file_logs=file_logs,
        config=conf_log,
    )
    if not isinstance(log, logging.Logger):
        raise Exception(log)

    # Copy URL files and write to S3 bucket or local machine
    is_repeat = True
    while is_repeat:

        # Do not repeat iteration if repeat_minutes is None
        is_repeat = repeat_minutes is not None

        start_time = time.time()

        log.info("******* START WEBSNAP ITERATION *******")
        log.info(
            f"Read config file: '{config}', it has sections: "
            f"{conf_parser.sections()}"
        )

        if s3_uploader:
            try:
                conf_s3 = validate_s3_config()
            except Exception as e:
                raise e
            write_urls_to_s3(
                conf_parser,
                conf_s3,
                log,
                min_size_kb,
                backup_s3_count,
                timeout,
                early_exit,
            )
        else:
            write_urls_locally(conf_parser, log, min_size_kb, timeout, early_exit)

        log.info("Finished websnap iteration")

        if is_repeat:  # pragma: no cover
            sleep_until_next_iteration(repeat_minutes, start_time, log)

    return



"""Config utilities, parses and validates config .ini files."""

import configparser
import json
import os
from pathlib import Path

import requests
from pydantic import (
    BaseModel,
    ValidationError,
    PositiveInt,
    AnyHttpUrl,
    AnyUrl,
    field_validator,
    NonNegativeInt,
    TypeAdapter,
)
from typing import Optional, Any
from dotenv import load_dotenv

from websnap.constants import LogRotation, MIN_SIZE_KB, TIMEOUT


def validate_positive_integer(x: Any) -> int:
    """
    Return x if it is a positive integer.

    Args:
        x: The input value.
    """
    ta = TypeAdapter(PositiveInt)

    try:
        ta.validate_python(x)
        return x
    except ValidationError:
        raise Exception(f"{x} is not a a positive integer")


def validate_positive_int_args(
    timeout: int, backup_s3_count: int | None = None, repeat_minutes: int | None = None
) -> None:
    """
    Return None if validation passes (arguments are positive integers).
    If validation fails then raises Exception.
    None values are allowed for arguments backup_s3_count and repeat_minutes
    (validation still passes).

    Args:
        timeout: Number of seconds to wait for response for each HTTP request.
            If integer passed then it must be a positive integer.
        backup_s3_count: Copy and backup S3 objects in each config section
            <backup_s3_count> times,
            remove object with the oldest last modified timestamp.
            If integer passed then it must be a positive integer.
        repeat_minutes: Run websnap continuously every <repeat> minutes
               If integer passed then it must be a positive integer.
    """
    param_arg_dict = {
        "timeout": timeout,
        "backup_s3_count": backup_s3_count,
        "repeat_minutes": repeat_minutes,
    }

    param = None
    try:
        for param, arg in param_arg_dict.items():
            if param == "timeout":
                validate_positive_integer(arg)
            elif arg is not None:
                validate_positive_integer(arg)
    except Exception as e:
        raise Exception(f"Invalid argument passed for parameter {param}: {e}")

    return


def is_url(x: Any) -> bool:
    """
    Return True if x is a URL. Else return False.

     Args:
        x: The input value.
    """
    ta = TypeAdapter(AnyUrl)
    try:
        ta.validate_python(x)
        return True
    except ValidationError:
        return False


def merge_config_parsers(
    config_1: configparser.ConfigParser, config_2: configparser.ConfigParser
) -> configparser.ConfigParser:
    """
    Merges config_2 into config_1 and then return config_1.
    If sections or keys in config_2 exist in config_1,
    the values from config_2 will overwrite those in config_1.
    """
    for section in config_2.sections():

        if not config_1.has_section(section):
            config_1.add_section(section)

        for option, value in config_2.items(section):
            config_1.set(section, option, value)

    return config_1


def get_json_config_parser(config_path: Path) -> configparser.ConfigParser:
    """
    Returns ConfigParser instance with items read from JSON config file.

    Args:
        config_path: Path object to the .json config file.
    """
    try:
        with open(config_path, "r") as config_file:
            data = json.load(config_file)

        config_parser = configparser.ConfigParser()
        config_parser.read_dict(data)

        return config_parser

    except FileNotFoundError:
        raise Exception(f"File '{config_path}' not found")
    except Exception as e:  # pragma: no cover
        raise Exception(e)


def get_url_json_config_parser(
    config_url: str, timeout: int = TIMEOUT
) -> configparser.ConfigParser:
    """
    Returns ConfigParser instance with items read from JSON config URL.

    Args:
        config_url: URL with additional configuration sections.
        timeout: Number of seconds to wait for response for each HTTP request.
    """
    try:
        response = requests.get(config_url, timeout=timeout)

        if not response.ok:
            raise Exception(
                f"URL {config_url} returned unsuccessful "
                f"status code {response.status_code}"
            )

        data = response.json()

        config_parser = configparser.ConfigParser()
        config_parser.read_dict(data)

        return config_parser

    except requests.exceptions.Timeout:  # pragma: no cover
        raise Exception(
            f"URL {config_url} timed out while waiting {timeout} seconds for response"
        )
    except Exception as e:  # pragma: no cover
        raise Exception(e)


def get_json_section_config_parser(
    section_config: str, timeout: int = TIMEOUT
) -> configparser.ConfigParser:
    """
    Returns ConfigParser instance with items read from JSON section config file.
    Cannot be used to assign DEFAULT section in returned ConfigParser instance.

    Args:
        section_config: File or URL with additional configuration sections.
        timeout: Number of seconds to wait for response for each HTTP request.
    """
    try:

        if is_url(section_config):
            section_parser = get_url_json_config_parser(section_config, timeout)
        else:
            if (section_path := Path(section_config)).suffix == ".json":
                section_parser = get_json_config_parser(section_path)
            else:
                raise Exception("Section config extension must be '.json'")

        if not isinstance(section_parser, configparser.ConfigParser):
            raise Exception(section_parser)

        if section_parser.defaults():
            raise Exception(f"Section config cannot have a 'DEFAULT' section")

        return section_parser

    except Exception as e:  # pragma: no cover
        raise Exception(e)


def get_config_parser(
    config: str, section_config: str | None = None, timeout: int = TIMEOUT
) -> configparser.ConfigParser:
    """
    Return ConfigParser object.
    If section_config passed then merges config and section_config
    into one ConfigParser instance.

    Args:
        config: Path to .ini or .json configuration file.
        section_config (str): File or URL to obtain additional configuration sections.
                              Default value is None.
        timeout: Number of seconds to wait for response for each HTTP request.
    """
    try:
        conf_path = Path(config)

        if section_config and conf_path.suffix != ".json":
            raise Exception(
                f"Config '{config}' extension must be '.json' to also use "
                f"optional section config '{section_config}'"
            )
        elif conf_path.suffix == ".json":
            config_parser = get_json_config_parser(conf_path)
            if section_config:
                section_parser = get_json_section_config_parser(section_config, timeout)
                config_parser = merge_config_parsers(config_parser, section_parser)
        else:  # pragma: no cover
            config_parser = configparser.ConfigParser()
            conf = config_parser.read(conf_path)
            if not conf:
                raise Exception(f"File '{config}' not found")

        if len(config_parser.sections()) < 1:  # pragma: no cover
            raise Exception(f"File '{config}' does not have any sections")

        return config_parser

    except Exception as e:  # pragma: no cover
        raise Exception(e)


class LogConfigModel(BaseModel):
    """
    Class with required log config values and their types.
    """

    log_when: str
    log_interval: PositiveInt
    log_backup_count: NonNegativeInt


def validate_log_config(
    config_parser: configparser.ConfigParser,
) -> LogConfigModel:
    """
    Return LogConfigModel object.
    Returns Exception if parsing fails.

    Args:
        config_parser (configparser.ConfigParser): ConfigParser object
    """
    try:
        log = {
            "log_when": config_parser.get(
                "DEFAULT", "log_when", fallback=LogRotation.WHEN.value
            ),
            "log_interval": config_parser.getint(
                "DEFAULT", "log_interval", fallback=LogRotation.INTERVAL.value
            ),
            "log_backup_count": config_parser.getint(
                "DEFAULT", "log_backup_count", fallback=LogRotation.BACKUP_COUNT.value
            ),
        }
        return LogConfigModel(**log)
    except ValidationError as e:
        raise Exception(f"Failed to validate config, error(s): {e}")
    except ValueError as e:
        raise Exception(f"Incorrect log related value in config, error(s): {e}")
    except Exception as e:
        raise Exception(f"{e}")


def validate_min_size_kb(config_parser: configparser.ConfigParser) -> int:
    """
    Return min_size_kb from config as integer.

    Args:
        config_parser: ConfigParser object
    """
    try:
        min_size_kb = config_parser.getint(
            "DEFAULT", "min_size_kb", fallback=MIN_SIZE_KB
        )
        if min_size_kb >= 0:
            return min_size_kb
        else:
            raise ValueError(
                "Value for config value 'min_size_kb' must be greater than or equal "
                "to 0"
            )
    except ValidationError as e:
        raise Exception(f"Failed to validate config value 'min_size_kb, error(s): {e}")
    except ValueError as e:  # pragma: no cover
        raise Exception(
            f"Incorrect value for config value 'min_size_kb', error(s): {e}"
        )
    except Exception as e:  # pragma: no cover
        raise Exception(f"{e}")


class ConfigSectionModel(BaseModel):
    """
    Class with required config section values (for writing to local machine).
    """

    url: AnyHttpUrl
    file_name: str
    directory: Optional[str] = None


def validate_config_section(
    config_parser: configparser.ConfigParser, section: str
) -> ConfigSectionModel | Exception:
    """
    Return ConfigSectionModel object.
    Returns Exception if parsing fails.

    Args:
        config_parser: ConfigParser object
        section: Name of section being validated
    """
    try:
        conf_section = {
            "url": config_parser.get(section, "url"),
            "file_name": config_parser.get(section, "file_name"),
        }
        if directory := config_parser.get(section, "directory", fallback=None):
            conf_section["directory"] = directory
        return ConfigSectionModel(**conf_section)
    except ValidationError as e:
        return Exception(
            f"Failed to validate config section '{section}', error(s): {e}"
        )
    except ValueError as e:
        return Exception(
            f"Incorrect value in config section '{section}', error(s): {e}"
        )
    except Exception as e:
        return Exception(f"{e}")


class S3ConfigModel(BaseModel):
    """
    Class with required S3 config values and their types.
    """

    endpoint_url: AnyUrl
    aws_access_key_id: str
    aws_secret_access_key: str


def validate_s3_config() -> S3ConfigModel:
    """
    Return S3ConfigModel object after validating required environment variables.
    """
    try:
        load_dotenv()
        s3_conf = {
            "endpoint_url": os.getenv("ENDPOINT_URL"),
            "aws_access_key_id": os.getenv("AWS_ACCESS_KEY_ID"),
            "aws_secret_access_key": os.getenv("AWS_SECRET_ACCESS_KEY"),
        }
        return S3ConfigModel(**s3_conf)
    except ValidationError as e:
        raise Exception(
            f"Failed to validate S3 config environment variables, error(s): {e}"
        )
    except Exception as e:  # pragma: no cover
        raise Exception(e)


class S3ConfigSectionModel(BaseModel):
    """
    Class with required config section values (for writing to S3 bucket).
    """

    url: AnyUrl
    bucket: str
    key: str

    @field_validator("key")
    @classmethod
    def key_must_contain_period(cls, v: str) -> str:
        key_split = v.rpartition(".")
        if not key_split[1]:
            raise ValueError("Config section key requires a file extension")
        return v

    @field_validator("key")
    @classmethod
    def key_must_not_start_with_slash(cls, v: str) -> str:
        if v.startswith("/"):
            raise ValueError("Config section key cannot start with a '/'")
        return v


def validate_s3_config_section(
    config_parser: configparser.ConfigParser, section: str
) -> S3ConfigSectionModel | Exception:
    """
    Return S3ConfigSectionModel object.
    Returns Exception if parsing fails.

    Args:
        config_parser: ConfigParser object
        section: Name of section being validated
    """
    try:
        conf_section = {
            "url": config_parser.get(section, "url"),
            "bucket": config_parser.get(section, "bucket"),
            "key": config_parser.get(section, "key"),
        }
        return S3ConfigSectionModel(**conf_section)
    except ValidationError as e:
        return Exception(
            f"Failed to validate config section '{section}', error(s): {e}"
        )
    except Exception as e:
        return Exception(f"{e}")



"""Tests for src/websnap/logger.py"""

import logging
import pytest

from websnap.logger import (
    get_custom_logger,
    get_log_level,
    get_console_handler,
    get_file_handler,
)


def test_get_custom_logger(log_config_model):

    log = get_custom_logger(
        name="websnap_logger", config=log_config_model, file_logs=True
    )
    assert isinstance(log, logging.Logger)

    file_handler_exists = False
    for handler in log.handlers:
        if isinstance(handler, logging.FileHandler):
            file_handler_exists = True

    assert file_handler_exists


@pytest.mark.parametrize(
    "log_level, expected",
    [
        ("DEBUG", 10),
        ("INFO", 20),
        ("WARNING", 30),
        ("ERROR", 40),
        ("CRITICAL", 50),
        ("random_string", 20),
    ],
)
def test_get_log_level(log_level, expected):
    assert get_log_level(log_level) == expected


def test_get_console_handler():
    assert isinstance(get_console_handler(), logging.StreamHandler)


@pytest.mark.parametrize(
    "filename, when, interval, backup",
    [("websnap1.log", "D", 1, 1), ("websnap2.log", "midnight", 2, 3)],
)
def test_get_log_config(filename, when, interval, backup):
    file_handler = get_file_handler(filename, when, interval, backup)
    assert isinstance(file_handler, logging.FileHandler)



"""Tests for src/websnap/websnap.py"""

import json
import os
import pytest
import requests
from dotenv import load_dotenv

from websnap import websnap


def test_websnap(config_basic, config_min_size_kb, config_log, tmp_path):

    for conf in [config_basic, config_min_size_kb, config_log]:

        config_path, tmp_path, file_name = conf

        websnap(config=config_path, early_exit=True)

        output_path = f"{str(tmp_path)}/{file_name}"

        assert os.path.isfile(output_path)
        assert os.path.getsize(output_path) > 999

        with open(output_path, "r") as f:
            data = json.load(f)
            assert data["info"]["name"] == "websnap"


# This test only supports S3 configurations for buckets with public read access
def test_websnap_s3(s3_config):

    if not s3_config:
        pytest.skip("Option '--s3-config' is not set")

    websnap(config=s3_config, s3_uploader=True, backup_s3_count=1, early_exit=True)

    with open(s3_config, "r") as f:
        s3_config_dict = json.load(f)

    load_dotenv()
    endpoint_url = os.getenv("ENDPOINT_URL")

    for section in s3_config_dict:

        if section == "DEFAULT":
            continue

        bucket = s3_config_dict[section]["bucket"]
        key = s3_config_dict[section]["key"]

        output_url = f"{endpoint_url}/{bucket}/{key}"

        response = requests.get(output_url, timeout=30)
        assert response.status_code == 200



"""
CLI that supports copying files hosted at URLs in config and then uploading
them to S3 bucket or local machine.

Example pdm command without flags (uses default argument values):
    pdm run websnap-cli

Example pdm command, writes files locally and repeats every 60 minutes (1 hour):
pdm run websnap-cli --file_logs --repeat_minutes 60

Example pdm command, uploads files to a S3 bucket and
repeats every 1440 minutes (24 hours):
pdm run websnap-cli --file_logs --s3_uploader --backup_s3_count 3 --repeat_minutes 1440

Example command to run command directly with python
from project root directory without flags:
    python -m src.websnap.websnap_cli
"""

import argparse
import websnap
from websnap.constants import TIMEOUT


def parse_arguments() -> argparse.Namespace | None:
    """
    Parses command line arguments and return arguments as argparse.Namespace object.
    If parsing fails then return None.
    """

    parser = argparse.ArgumentParser(
        description="Supports copying files hosted at URLs in "
        "config and then uploading them to S3 bucket or local machine."
    )

    parser.add_argument(
        "-c",
        "--config",
        default="config.ini",
        help="Path to configuration file." "Default value is 'config.ini'.",
    )

    parser.add_argument(
        "-l",
        "--log_level",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        default="INFO",
        help="Level to use for logging. Default value is 'INFO'.",
    )

    parser.add_argument(
        "-f", "--file_logs", action="store_true", help="Enable rotating file logs."
    )

    parser.add_argument(
        "-s",
        "--s3_uploader",
        action="store_true",
        help="Enable uploading of files to S3 bucket. ",
    )

    parser.add_argument(
        "-b",
        "--backup_s3_count",
        type=int,
        help="Copy and backup S3 objects in each config section"
        "<backup_s3_count> times, "
        "remove object with the oldest last modified timestamp. "
        "If omitted then objects are not copied or removed.",
    )

    parser.add_argument(
        "-t",
        "--timeout",
        type=int,
        default=TIMEOUT,
        help="Number of seconds to wait for response for each HTTP request.",
    )

    parser.add_argument(
        "-e",
        "--early_exit",
        action="store_true",
        help="Enable early program termination after error occurs. "
        "If omitted then logs URL processing errors "
        "but continues program execution.",
    )

    parser.add_argument(
        "-r",
        "--repeat_minutes",
        type=int,
        help="Run websnap continuously every <repeat_minutes>. "
        "If omitted then websnap does not repeat.",
    )

    parser.add_argument(
        "-n",
        "--section_config",
        help="File or URL to obtain additional configuration sections. "
        "Cannot be used to assign DEFAULT section in config. "
        "Only currently supports JSON config.",
    )

    return parser.parse_args()


def main():
    """
    Main entry point for websnap-cli.
    """
    kwargs = vars(parse_arguments())

    websnap.websnap(
        config=kwargs["config"],
        log_level=kwargs["log_level"],
        file_logs=kwargs["file_logs"],
        s3_uploader=kwargs["s3_uploader"],
        backup_s3_count=kwargs["backup_s3_count"],
        timeout=kwargs["timeout"],
        early_exit=kwargs["early_exit"],
        repeat_minutes=kwargs["repeat_minutes"],
        section_config=kwargs["section_config"],
    )



"""Tests for src/websnap/websnap_cli.py"""

import subprocess


def test_websnap_cli(config_basic):

    result_pass = subprocess.run(
        [
            "websnap_cli",
            f"--config={config_basic[0]}",
            "--log_level=WARNING",
            "--file_logs",
            "--timeout=30",
            "--early_exit",
        ],
        capture_output=True,
        text=True,
    )
    assert result_pass.returncode == 0

    result_fail = subprocess.run(
        [
            "websnap_cli",
            "--timeout=not_a_number",
        ],
        capture_output=True,
        text=True,
    )
    assert result_fail.returncode != 0



"""Tests for src/websnap/validators.py"""

import configparser
from pathlib import Path

import pytest

from websnap.validators import (
    is_url,
    merge_config_parsers,
    get_json_config_parser,
    get_url_json_config_parser,
    get_json_section_config_parser,
    get_config_parser,
    validate_positive_integer,
    validate_s3_config_section,
    S3ConfigSectionModel,
)


@pytest.mark.parametrize("x", [1, 0, 1.23, "abc"])
def test_validate_positive_integer(x):
    try:
        result = validate_positive_integer(x)
        assert result == x
    except Exception as e:
        if e == f"{x} is not a positive integer":
            assert True


@pytest.mark.parametrize(
    "x, expected",
    [
        ("https://pypi.org/pypi/websnap/json", True),
        ("A most lovely string", False),
        (1, False),
        (True, False),
        (None, False),
    ],
)
def test_is_url(x, expected):
    result = is_url(x)
    assert result == expected


def test_merge_config_parsers(config_parser_basic, config_parser_log, config_parser_s3):
    result_1 = merge_config_parsers(config_parser_basic, config_parser_log)
    assert isinstance(result_1, configparser.ConfigParser)

    result_2 = merge_config_parsers(config_parser_basic, config_parser_s3)
    assert isinstance(result_2, configparser.ConfigParser)


def test_json_config_parser(config_basic):
    result = get_json_config_parser(config_basic[0])
    assert isinstance(result, configparser.ConfigParser)


def test_json_config_parser_nonexistent_config():
    with pytest.raises(Exception):
        get_json_config_parser(Path("nonexistent_config.json"))


def test_get_url_json_config_parser():
    result_1 = get_url_json_config_parser(
        "https://www.envidat.ch/converters-api/internal-dataset/websnap-config-all/"
        "bibtex?bucket=random&is-recent=true&is-json=true",
        30,
    )
    assert isinstance(result_1, configparser.ConfigParser)

    with pytest.raises(Exception):
        get_url_json_config_parser("https://httpbin.org/status/400", 30)


def test_get_json_section_config_parser():
    result = get_json_section_config_parser(
        "https://www.envidat.ch/converters-api/internal-dataset/websnap-config-all/"
        "bibtex?bucket=random&is-recent=true&is-json=true",
        30,
    )
    assert isinstance(result, configparser.ConfigParser)


@pytest.mark.parametrize(
    "section_config", ["section_config.ini", "section_config.json"]
)
def test_get_json_section_config_parser_invalid_section_config(section_config):
    with pytest.raises(Exception):
        get_json_section_config_parser(section_config)


def test_get_config_parser(config_basic):
    result = get_config_parser(config=config_basic[0], timeout=30)
    assert isinstance(result, configparser.ConfigParser)


@pytest.mark.parametrize(
    "config, section_config, timeout",
    [
        ("config_1.ini", "section_config.ini", 30),
        ("config_2.json", "section_config.json", 30),
    ],
)
def test_get_config_parser_invalid_parameters(config, section_config, timeout):
    with pytest.raises(Exception):
        get_config_parser(config=config, section_config=section_config, timeout=timeout)


def test_get_config_parser_invalid_section_config(config_basic):
    with pytest.raises(Exception):
        get_config_parser(
            config=config_basic[0], section_config="non-existent.json", timeout=30
        )


def test_validate_s3_config_section(config_parser_s3):
    result = validate_s3_config_section(config_parser_s3, "pypi-websnap-s3")
    assert isinstance(result, S3ConfigSectionModel)


@pytest.mark.parametrize(
    "section",
    ["pypi-websnap-s3_invalid_key", "pypi-websnap-s3_invalid_key2", "no-bucket"],
)
def test_validate_s3_config_section_invalid_section(section, config_parser_s3_invalid):
    result = validate_s3_config_section(config_parser_s3_invalid, section)
    assert not isinstance(result, S3ConfigSectionModel)



"""
Functions used to copy and write files retrieved from an API to a S3 bucket or a
local machine.
"""

import configparser
import logging
import os
import time
from math import ceil

import requests
import boto3
from botocore.exceptions import ClientError
import sys

from websnap.validators import (
    validate_config_section,
    S3ConfigModel,
    validate_s3_config_section,
    ConfigSectionModel,
    S3ConfigSectionModel,
)


def terminate_program(early_exit: bool):
    """Terminates program execution if argument early_exit is True."""
    if early_exit:
        sys.exit("Error occurred: check logs for details")
    return


def get_url_content(
    url: str,
    section: str,
    log: logging.getLogger,
    timeout: int = 32,
    early_exit: bool = False,
) -> bytes | None:
    """
    Return content of response from HTTP GET request.
    If response times out or response status code is >= 400 then terminate program if
    argument early_exit is True, else return None.

    Args:
        url: URL to download.
        section: Name of config section being processed.
        log: Logger object created with customized configuration file.
        timeout: Number of seconds to wait for response.
        early_exit: If True then terminates program immediately after error occurs.
            Default value is False.
            If False then only logs error and continues execution.
    """
    try:

        response = requests.get(url, timeout=timeout)

        if not response.ok:
            log.error(
                f"Config section '{section}': "
                f"URL returned unsuccessful HTTP response "
                f"status code {response.status_code}"
            )
            terminate_program(early_exit)
            return None

        return response.content

    except requests.exceptions.Timeout:  # pragma: no cover
        log.error(
            f"Config section '{section}': "
            f"URL timed out while waiting {timeout} seconds for response"
        )
        terminate_program(early_exit)
        return None


def is_min_size_kb(
    url_content: bytes,
    min_size_kb: int,
    section: str,
    log: logging.getLogger,
    early_exit: bool = False,
) -> bool | None:
    """
    Return True if url_content is greater than min_size_kb.
    Else return False or terminate program (if argument early_exit is True).

    Args:
        url_content: Content of response from HTTP request.
        min_size_kb: Minimum threshold in kilobytes that URL response content must be to
            write or upload file.
        section: Name of config section being processed.
        log: Logger object created with customized configuration file.
        early_exit: If True then terminates program immediately after error occurs.
            Default value is False.
            If False then only logs error and continues execution.
    """
    data_kb = url_content.__sizeof__() / 1024

    if data_kb < min_size_kb:
        log.error(
            f"Config section '{section}': "
            f"URL response content in config section {section} is less than "
            f"config value 'min_size_kb' {min_size_kb}"
        )
        terminate_program(early_exit)
        return False

    return True


def write_urls_locally(
    conf_parser: configparser.ConfigParser,
    log: logging.getLogger,
    min_size_kb: int,
    timeout: int = 32,
    early_exit: bool = False,
) -> None:
    """
    Download files hosted at URLS in config and then write them to local machine.

    Args:
        conf_parser: ConfigParser object created from parsing configuration file.
        log: Logger object created with customized configuration file.
        min_size_kb: Minimum threshold in kilobytes that URL response content must be to
            write file.
        timeout: Number of seconds to wait for response.
        early_exit: If True then terminates program immediately after error occurs.
            Default value is False.
            If False then only logs error and continues execution.
    """
    for section in conf_parser.sections():

        try:
            conf = validate_config_section(conf_parser, section)
            if not isinstance(conf, ConfigSectionModel):
                log.error(f"Config section '{section}': {conf}")
                terminate_program(early_exit)
                continue

            if conf.directory and not os.path.isdir(conf.directory):  # pragma: no cover
                log.error(
                    f"Config section '{section}': directory '{conf.directory}' "
                    f"does not exist"
                )
                terminate_program(early_exit)
                continue

            url_content = get_url_content(
                str(conf.url), section, log, timeout, early_exit
            )
            if not url_content:  # pragma: no cover
                continue

            is_min_size = is_min_size_kb(
                url_content, min_size_kb, section, log, early_exit
            )
            if not is_min_size:  # pragma: no cover
                continue

            if conf.directory:
                file_path = f"{conf.directory}/{conf.file_name}"
            else:  # pragma: no cover
                file_path = f"{conf.file_name}"

            with open(file_path, "wb") as f:
                f.write(url_content)
                log.info(
                    f"Successfully downloaded URL content and wrote file locally in "
                    f"config section: {section}"
                )

        except Exception as e:  # pragma: no cover
            log.error(f"Config section '{section}', error(s): {e}")
            terminate_program(early_exit)

    return


def handle_s3_client_error(
    err: ClientError, log: logging.getLogger, section: str, early_exit: bool
) -> None:  # pragma: no cover
    """
    Handles and logs botocore.exceptions.ClientError returned by failed S3 client
    method call.

    Note: If HTTP status code 404 is returned with ClientError then logs warning but
    continues execution.

    Args:
        err: Client error returned from S3 client.
        log: Logger object created with customized configuration file.
        section: Name of config section being processed.
        early_exit: If True then terminates program immediately after error occurs.
            If False then only logs error and continues execution.
    """
    err_status_code = err.response.get("ResponseMetadata", {}).get("HTTPStatusCode")

    if err_status_code == 404:
        log.warning(f"Config section '{section}': Object not found")
    elif err_status_code == 403:
        log.error(
            f"Config section '{section}': Forbidden, check access credentials in config"
        )
        terminate_program(early_exit)
    else:
        log.error(err)
        terminate_program(early_exit)

    return


def copy_s3_object(
    client: boto3.Session.client,
    conf: S3ConfigSectionModel,
    log: logging.getLogger,
    section: str,
    early_exit: bool = False,
) -> None:
    """
    Copy an object using S3 object config.

    New object's name is constructed using the 'LastModified' timestamp of original
    object.

    Args:
        client : boto3.Session.client object created using configuration file values.
        conf: S3ConfigSectionModel object created from validated
            section of configuration file.
        log: Logger object created with customized configuration file.
        section: Name of config section being processed.
        early_exit: If True then terminates program immediately after error occurs.
            Default value is False.
            If False then only logs error and continues execution.
    """

    try:
        head_resp = client.head_object(Bucket=conf.bucket, Key=conf.key)

        last_modified = head_resp.get("LastModified")
        format_date = "%Y-%m-%d_%H-%M-%S"
        datetime_str = last_modified.strftime(format_date)
        key_split = conf.key.rpartition(".")
        key_copy = f"{key_split[0]}_{datetime_str}{key_split[1]}{key_split[2]}"

        response = client.copy_object(
            CopySource={"Bucket": conf.bucket, "Key": conf.key},
            Bucket=conf.bucket,
            Key=key_copy,
        )

        status_code = response.get("ResponseMetadata", {}).get("HTTPStatusCode")

        if status_code == 200:
            log.info(
                f"Config section '{section}': " f"Created new backup file '{key_copy}'"
            )
        else:  # pragma: no cover
            log.error(
                f"Config section '{section}': "
                f"Object backup attempt returned "
                f"unexpected HTTP response {status_code}"
            )
            terminate_program(early_exit)

    except ClientError as err:
        handle_s3_client_error(err, log, section, early_exit)

    return


def delete_s3_backup_object(
    client: boto3.Session.client,
    conf: S3ConfigSectionModel,
    log: logging.getLogger,
    section: str,
    backup_s3_count: int,
    early_exit: bool = False,
) -> None:
    """
    Delete a S3 backup object using S3 object config.
    Only deletes object if backup objects exceed backup_s3_count.

    Only deletes object that corresponds to the file name in the configured key,
    allows for a timestamp in key created using copy_s3_object().

    Args:
        client : boto3.Session.client object created using configuration file values.
        conf: S3ConfigSectionModel object created from validated
            section of configuration file.
        log: Logger object created with customized configuration file.
        section: Name of config section being processed.
        backup_s3_count: Copy and backup S3 objects in config
            <backup_s3_count> times, remove object with the oldest last modified
            timestamp.
        early_exit: If True then terminates program immediately after error occurs.
            Default value is False.
            If False then only logs error and continues execution.
    """

    try:
        key_split = conf.key.rpartition("/")

        if not key_split[0]:
            response = client.list_objects_v2(  # pragma: no cover
                Bucket=conf.bucket,
            )
        else:
            response = client.list_objects_v2(
                Bucket=conf.bucket, Prefix=f"{key_split[0]}{key_split[1]}"
            )

        file_split = key_split[2].rpartition(".")
        file_start = f"{file_split[0]}_"
        file_end = f"{file_split[1]}{file_split[2]}"

        objs = [obj for obj in response.get("Contents")]
        match_objs = []

        for obj in objs:
            ky = obj.get("Key")
            ky_split = ky.rpartition("/")
            ky_file = ky_split[2]
            if ky_file.startswith(file_start) and ky_file.endswith(file_end):
                match_objs.append(obj)

        sorted_objs = sorted(match_objs, key=lambda x: x["LastModified"])

        if len(sorted_objs) > backup_s3_count:

            obj_oldest = sorted_objs[0]
            delete_key = obj_oldest.get("Key")

            resp = client.delete_object(Bucket=conf.bucket, Key=delete_key)

            status_code = resp.get("ResponseMetadata", {}).get("HTTPStatusCode")

            if status_code == 204:
                log.info(
                    f"Config section '{section}': "
                    f"Deleted backup file '{delete_key}'"
                )
            else:  # pragma: no cover
                log.error(
                    f"Config section '{section}': Backup file delete "
                    f"attempt returned unexpected HTTP response {status_code}"
                )
                terminate_program(early_exit)

        else:  # pragma: no cover
            log.info(
                f"Config section '{section}': Current number of backup "
                f"files does not exceed backup S3 count {backup_s3_count}"
            )

    except ClientError as err:
        handle_s3_client_error(err, log, section, early_exit)

    return


def write_urls_to_s3(
    conf_parser: configparser.ConfigParser,
    conf_s3: S3ConfigModel,
    log: logging.getLogger,
    min_size_kb: int,
    backup_s3_count: int | None = None,
    timeout: int = 32,
    early_exit: bool = False,
) -> None:
    """
    Download files hosted at URLS in config and then upload them to S3 bucket.

    Args:
        conf_parser: ConfigParser object created from parsing configuration file.
        conf_s3: S3ConfigModel object created from validated configuration file.
        log: Logger object created with customized configuration file.
        min_size_kb: Minimum threshold in kilobytes that URL response content must be to
            upload file to S3 bucket.
        backup_s3_count: Copy and backup S3 objects in each config section
            <backup_s3_count> times,
            remove object with the oldest last modified timestamp.
            If omitted then default value is None and objects are not copied or removed.
        timeout: Number of seconds to wait for response.
        early_exit: If True then terminates program immediately after error occurs.
            Default value is False.
            If False then only logs error and continues execution.
    """
    session = boto3.Session(
        aws_access_key_id=conf_s3.aws_access_key_id,
        aws_secret_access_key=conf_s3.aws_secret_access_key,
    )

    client = session.client(service_name="s3", endpoint_url=str(conf_s3.endpoint_url))

    for section in conf_parser.sections():

        try:
            conf = validate_s3_config_section(conf_parser, section)
            if not isinstance(conf, S3ConfigSectionModel):  # pragma: no cover
                log.error(f"Config section '{section}': {conf}")
                terminate_program(early_exit)
                continue

            url_content = get_url_content(
                str(conf.url), section, log, timeout, early_exit
            )
            if not url_content:  # pragma: no cover
                continue

            is_min_size = is_min_size_kb(
                url_content, min_size_kb, section, log, early_exit
            )
            if not is_min_size:  # pragma: no cover
                continue

            if backup_s3_count:
                copy_s3_object(client, conf, log, section, early_exit)
                delete_s3_backup_object(
                    client, conf, log, section, backup_s3_count, early_exit
                )

            response_s3 = client.put_object(
                Body=url_content, Bucket=conf.bucket, Key=conf.key
            )
            status_code = response_s3.get("ResponseMetadata", {}).get("HTTPStatusCode")

            if status_code == 200:
                log.info(
                    f"Config section '{section}': Successfully copied URL "
                    f"content to S3 object '{conf.key}'"
                )
            else:  # pragma: no cover
                log.error(
                    f"Config section '{section}': S3 client returned unexpected "
                    f"HTTP response {status_code}"
                )
                terminate_program(early_exit)

        except ClientError as err:
            handle_s3_client_error(err, log, section, early_exit)

        except Exception as e:  # pragma: no cover
            log.error(f"Config section '{section}', error(s): {e}")
            terminate_program(early_exit)

    return


def sleep_until_next_iteration(
    sleep_minutes: int, start_time: float, log: logging.getLogger
) -> None:  # pragma: no cover
    """
    Sleep (delay execution) <sleep_minutes> minutes until next websnap iteration.

    Args:
        sleep_minutes: Number of minutes to sleep until executing next websnap
            iteration.
        start_time: Start time at beginning of websnap iteration,
            (time is specified as number of seconds since beginning of epoch.)
        log: Logger object created with customized configuration file.
    """
    sleep_secs = sleep_minutes * 60
    execution_secs = time.time() - start_time

    if sleep_secs > execution_secs:
        wait_seconds = int(sleep_secs - execution_secs)
        wait_minutes = wait_seconds / 60
        log.info(
            f"Sleeping {wait_seconds} seconds (about {ceil(wait_minutes)} minutes)"
            f" before next iteration..."
        )
        time.sleep(wait_seconds)

    return



"""Helpers for tests."""

import json


def write_json_config(config_path: str, conf_dict: dict):
    with open(config_path, "w") as f:
        f.write(json.dumps(conf_dict))


def get_section_config(tmp_path, file_name: str):
    return {
        "pypi-websnap": {
            "directory": str(tmp_path),
            "file_name": file_name,
            "url": "https://pypi.org/pypi/websnap/json",
        },
    }


def get_s3_config():
    return {
        "pypi-websnap-s3": {
            "url": "https://pypi.org/pypi/websnap/json",
            "bucket": "pypi",
            "key": "output_s3.json",
        },
    }


def get_s3_config_invalid():
    return {
        "pypi-websnap-s3_invalid_key": {
            "url": "https://pypi.org/pypi/websnap/json",
            "bucket": "pypi",
            "key": "/output_s3.json",
        },
        "pypi-websnap-s3_invalid_key2": {
            "url": "https://pypi.org/pypi/websnap/json",
            "bucket": "pypi",
            "key": "output_s3json",
        },
        "no-bucket": {
            "url": "https://pypi.org/pypi/websnap/json",
            "key": "output_s3json",
        },
    }
